@@ -25,7 +25,9 @@ module Bundler
 
             cache_uri = original_uri || uri
 
-            uri_parts = [cache_uri.host, cache_uri.user, cache_uri.port, cache_uri.path]
+            host = cache_uri.to_s.start_with?("file://") ? nil : cache_uri.host
+
+            uri_parts = [host, cache_uri.user, cache_uri.port, cache_uri.path]
             uri_digest = SharedHelpers.digest(:MD5).hexdigest(uri_parts.compact.join("."))
 
             uri_parts[-1] = uri_digest
@@ -46,7 +48,7 @@ module Bundler
           end
 
           uri
-        rescue URI::InvalidComponentError
+        rescue Bundler::URI::InvalidComponentError
           error_message = "Please CGI escape your usernames and passwords before " \
                           "setting them for authentication."
           raise HTTPError.new(error_message)
